/*
 * unity-webapps-context-daemon.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdlib.h>

#include "unity-webapps-context-daemon.h"
#include "../gen-webapps-context-daemon.xml.h"
#include "../gen-webapps-service.xml.h"

#include "../unity-webapps-dbus-defs.h"

static GMainLoop *mainloop = NULL;

static GDBusNodeInfo *context_introspection_data = NULL;
static GDBusNodeInfo *service_introspection_data = NULL;

static GDBusConnection *session_bus = NULL;

static guint object_registration;

typedef struct _UnityWebappsContextInfo {
  gchar *name;
  gchar *domain;
} UnityWebappsContextInfo;

static UnityWebappsContextInfo info;

static gchar *
get_canonical_full_name (const gchar *name, const gchar *domain)
{
  gchar *canonical_name, **components;


  canonical_name = g_strdup_printf ("%s%s", name, domain);
  g_strcanon (canonical_name, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUV\
WXYZ123567890", '%');

  components = g_strsplit (canonical_name, "%", 0);

  g_free (canonical_name);

  canonical_name = g_strjoinv (NULL, components);

  g_strfreev(components);

  return canonical_name;
}

/* Note we need to be more intelligent about this currently we wont write a new desktop file if we are given an icon when we didnt have one before */
static void
write_desktop_file ()
{
  gchar *path, *contents, *desktop_file_name, *t;
  
  t = get_canonical_full_name (info.name, info.domain);
  desktop_file_name = g_strdup_printf("%s.desktop", t);
  g_free (t);
  
  path = g_strdup_printf("%s/.data/applications/%s", g_get_home_dir(), desktop_file_name);
  
  contents = g_strdup_printf("[Desktop Entry]\nName=%s\nType=Application\nIcon=%s", info.name, "application-default-icon");
  
  g_file_set_contents (path, contents, -1, NULL);
  
  g_free (desktop_file_name);
  g_free (path);
  g_free (contents);

}

static gchar *
get_desktop_file_name ()
{
  gchar *path, *desktop_file_name, *t;

  t = get_canonical_full_name (info.name, info.domain);
  desktop_file_name = g_strdup_printf("%s.desktop", t);
  g_free (t);
  
  path = g_strdup_printf("%s/.data/applications/%s", g_get_home_dir(), desktop_file_name);
  
  if (g_file_test(path, G_FILE_TEST_EXISTS) == FALSE)
    {
      write_desktop_file();
    }
  g_free (path);
  
  return desktop_file_name;

}


static void
bus_filter_service_ready (GDBusConnection *connection,
			  const gchar *sender,
			  const gchar *object,
			  const gchar *interface,
			  const gchar *signal,
			  GVariant *params,
			  gpointer user_data)
{
  GError *error;
  UnityWebappsContextDaemon *daemon;
  
  daemon = (UnityWebappsContextDaemon *)user_data;

  if (g_strcmp0(signal, "ServiceReady") != 0)
    {
      g_warning ("Unexpected signal on context daemon instance: %s", signal);
      return;
    }

  error = NULL;
  g_dbus_connection_emit_signal (session_bus,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ContextReady",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_critical ("Error emiting ContextReady signal in ServiceReady: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

static void
interest_name_vanished (GDBusConnection *connection,
			const gchar *name,
			gpointer user_data)
{
  UnityWebappsContextDaemon *daemon;
  guint watcher_id;
  guint num_interests;
  
  daemon = (UnityWebappsContextDaemon *)user_data;
  
  num_interests = GPOINTER_TO_UINT (g_hash_table_lookup (daemon->name_interest_counts, name));
  
  watcher_id = GPOINTER_TO_UINT (g_hash_table_lookup (daemon->interest_vanished_ids,
						      name));
  if (watcher_id == 0 || num_interests == 0)
    {
      return;
    }
  else
    {
      g_bus_unwatch_name (watcher_id);
      g_hash_table_remove (daemon->interest_vanished_ids, name);
    }

  daemon->interests -= num_interests;

  g_hash_table_remove (daemon->name_interest_counts, name);
  
  if (daemon->interests == 0)
    {
      GError *error;
      
      error = NULL;
      g_dbus_connection_emit_signal (connection,
				     NULL,
				     UNITY_WEBAPPS_CONTEXT_PATH,
				     UNITY_WEBAPPS_CONTEXT_IFACE,
				     "NoInterest",
				     NULL, /* params */
				     &error /* Error */);
      
      if (error != NULL)
	{
	  g_critical ("Error emitting NoInterest signal: %s", error->message);
	  g_error_free (error);
	  return;
	}
    }


}

static void
add_interest (GDBusConnection *connection,
	      UnityWebappsContextDaemon *daemon,
	      const gchar *name)
{
  guint watcher_id;

  daemon->interests++;
  
  g_hash_table_insert (daemon->name_interest_counts, g_strdup (name),
		       GUINT_TO_POINTER 
		       (GPOINTER_TO_UINT 
			(g_hash_table_lookup (daemon->name_interest_counts, name))+1));
										
  
  watcher_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
				 name,
				 G_BUS_NAME_WATCHER_FLAGS_NONE,
				 NULL,
				 interest_name_vanished,
				 daemon,
				 NULL /* User data free */);
  
  g_hash_table_insert (daemon->interest_vanished_ids, g_strdup (name), (GUINT_TO_POINTER(watcher_id)));
}

static void
lost_interest (GDBusConnection *connection,
	       UnityWebappsContextDaemon *daemon,
	       const gchar *name)
{
  guint watcher_id;
  guint num_interests;
  
  watcher_id = GPOINTER_TO_UINT (g_hash_table_lookup (daemon->interest_vanished_ids, name));
  num_interests = GPOINTER_TO_UINT (g_hash_table_lookup (daemon->name_interest_counts, name));
  
  
  if (num_interests == 1)
    {
      if (watcher_id == 0)
	{
	  g_message("Lost interest from name we didn't expect: %s", name);
	  return;
	}
      g_bus_unwatch_name (watcher_id);
      
      g_hash_table_remove (daemon->interest_vanished_ids, name);
    }
  

  if (daemon->interests == 0)
    {
      g_warning ("Somehow we fell below 0 interests");
      return;
    }
  
  daemon->interests--;
  
  g_hash_table_insert (daemon->name_interest_counts, g_strdup (name),
		       GUINT_TO_POINTER (num_interests-1));

  if (daemon->interests == 0)
    {
      GError *error;
      
      error = NULL;
      g_dbus_connection_emit_signal (connection,
				     NULL,
				     UNITY_WEBAPPS_CONTEXT_PATH,
				     UNITY_WEBAPPS_CONTEXT_IFACE,
				     "NoInterest",
				     NULL, /* params */
				     &error /* Error */);
      
      if (error != NULL)
	{
	  g_critical ("Failed to emit NoInterest signal: %s", error->message);
	  g_error_free (error);
	  return;
	}

    }
}

static void
handle_method_call (GDBusConnection *connection,
		    const gchar *sender,
		    const gchar *object_path,
		    const gchar *interface_name,
		    const gchar *method_name,
		    GVariant *parameters,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UnityWebappsContextDaemon *daemon;
  
  daemon = (UnityWebappsContextDaemon *)user_data;
  
  if (g_strcmp0 (method_name, "GetName") == 0)
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", info.name));
    }
  else if (g_strcmp0(method_name, "GetDomain") == 0)
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", info.domain));
    }
  else if (g_strcmp0(method_name, "GetDesktopName") == 0)
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", daemon->desktop_file_name));
    }
  else if (g_strcmp0(method_name, "AddInterest") == 0)
    {
      const gchar *name = NULL;
      
      g_variant_get (parameters, "(s)", &name, NULL);
      
      add_interest(connection,daemon, name);
      
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else if (g_strcmp0(method_name, "LostInterest") == 0)
    {
      lost_interest(connection, daemon, sender);

      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else if (g_strcmp0(method_name, "Shutdown") == 0)
    {
      g_dbus_method_invocation_return_value (invocation, NULL);
      g_dbus_connection_flush_sync (connection, NULL, NULL);

      exit(0);
    }
  else
    {
      g_warning ("Unexpected method in context daemon instance: %s", method_name);
      return;
    }
}

static const GDBusInterfaceVTable interface_vtable = 
  {
    handle_method_call, 
    NULL,
    NULL
  };

static void
bus_get_cb (GObject *object, GAsyncResult *res, gpointer user_data)
{
  GError *error = NULL;
  GDBusConnection *connection = g_bus_get_finish(res, &error);
  UnityWebappsContextDaemon *daemon;
 
  
  if (error != NULL) 
    {
      g_error("Unable to get session bus: %s", error->message);
      g_error_free(error);
      return;
    }
  
  session_bus = connection;
  
  // TODO: Investigate free on shutdown, just for correctness.
  daemon = unity_webapps_context_daemon_new ();
  
  object_registration = 
    g_dbus_connection_register_object (session_bus,
				       UNITY_WEBAPPS_CONTEXT_PATH,
				       context_introspection_data->interfaces[0],
				       &interface_vtable, 
				       daemon /* User Data*/, 
				       NULL /* User Data free*/, 
				       &error);
  
  if (error != NULL) 
    {
      g_error ("Unable to register context on the bus: %s", error->message);

      g_error_free(error);
      unity_webapps_context_daemon_free (daemon);

      return;
    }
  
  g_dbus_connection_emit_signal (session_bus,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ContextReady",
				 NULL, /* params */
				 &error);
  
  if (error != NULL)
    {
      g_critical ("Unable to send the 'ContextReady' signal: %s", error->message);
      g_error_free (error);
      unity_webapps_context_daemon_free (daemon);

      return;
    }
  
  g_dbus_connection_signal_subscribe (session_bus,
				      NULL,
				      UNITY_WEBAPPS_SERVICE_IFACE,
				      "ServiceReady",
				      NULL,
				      NULL,
				      G_DBUS_SIGNAL_FLAGS_NONE,
				      bus_filter_service_ready,
				      daemon /* user data */,
				      NULL /* destroy notify */);
				 
}

static gboolean
parse_args (gint argc, gchar **argv)
{
  if (argc != 3)
    {
    return FALSE;
    }
  else
    {
      info.name = argv[1];
      info.domain = argv[2];
      
      return TRUE;
    }
}

gint 
main (gint argc, gchar **argv)
{
  GError *error;
  guint owner_id;

  g_type_init ();
  
  if (!parse_args (argc, argv))
    return 1;
  
  error = NULL;
  
  context_introspection_data = g_dbus_node_info_new_for_xml (_webapps_context_daemon, &error);
  
  if (error != NULL)
    {
      g_error ("Can not load context introspection data: %s", error->message);
      g_error_free (error);
      
      return 1;
    }
  
  service_introspection_data = g_dbus_node_info_new_for_xml (_webapps_service, NULL);

  
  if (error != NULL)
    {
      g_error ("Can not load service introspection data: %s", error->message);
      g_error_free (error);
      
      return 1;
    }
  
  g_bus_get(G_BUS_TYPE_SESSION,
	    NULL, /* Cancellable */
	    bus_get_cb,
	    NULL /* User data */);
  mainloop = g_main_loop_new(NULL, FALSE);
  g_main_loop_run(mainloop);
  
  g_dbus_node_info_unref (context_introspection_data);
  g_dbus_node_info_unref (service_introspection_data);

  return 0;
}


UnityWebappsContextDaemon *
unity_webapps_context_daemon_new ()
{
  UnityWebappsContextDaemon *daemon;
  
  daemon = g_malloc0 (sizeof (UnityWebappsContextDaemon));
  
  daemon->interests = 0;
  
  daemon->interest_vanished_ids = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  daemon->name_interest_counts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

  daemon->desktop_file_name = get_desktop_file_name ();
  
  daemon->notification_context = unity_webapps_notification_context_new (session_bus, info.name);
  daemon->indicator_context = unity_webapps_indicator_context_new (session_bus, daemon->desktop_file_name);
  daemon->music_player_context = unity_webapps_music_player_context_new (session_bus, daemon->desktop_file_name);
  

  
  return daemon;
}

void
unity_webapps_context_daemon_free (UnityWebappsContextDaemon *daemon)
{
  // TODO: Can we come back after this?
  g_hash_table_destroy (daemon->interest_vanished_ids);
  g_hash_table_destroy (daemon->name_interest_counts);
  
  unity_webapps_notification_context_free (daemon->notification_context);
  unity_webapps_indicator_context_free (daemon->indicator_context);
  unity_webapps_music_player_context_free (daemon->music_player_context);
  
  g_free (daemon->desktop_file_name);

  g_free (daemon);
}
