/*
 * unity-webapps-indicator-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include "unity-webapps-indicator-context.h"
#include "../gen-webapps-context-daemon.xml.h"
#include "../gen-webapps-service.xml.h"
#include "../gen-webapps-indicator.xml.h"

#include "../unity-webapps-dbus-defs.h"

static GDBusNodeInfo *indicator_introspection_data = NULL;

static void
real_clear_indicator (UnityWebappsIndicatorContext *indicator_context,
		      const gchar *name)
{
  IndicateIndicator *indicator;

  indicator = g_hash_table_lookup (indicator_context->indicators_by_name,
				   name);
  
  if (indicator == NULL)
    return;
  
  indicate_indicator_hide (indicator);
  
  g_hash_table_remove (indicator_context->indicators_by_name, name);
}

static void
real_set_property (UnityWebappsIndicatorContext *indicator_context,
		   const gchar *name,
		   const gchar *property,
		   const gchar *value)
{
  IndicateIndicator *indicator;
  
  indicator = g_hash_table_lookup (indicator_context->indicators_by_name, name);
  
  if (indicator == NULL)
    {
      return;
    }
  
  indicate_indicator_set_property (indicator, property, value);
}

static void
real_show_indicator (UnityWebappsIndicatorContext *indicator_context,
		const gchar *name)
{
  IndicateIndicator *indicator;
  
  if ((indicator =g_hash_table_lookup (indicator_context->indicators_by_name, name)) != NULL)
    {
      indicate_indicator_show (indicator);
    }
  
  indicator = indicate_indicator_new_with_server (indicator_context->indicate_server);
  indicate_indicator_set_property (indicator, "name", name);
  
  indicate_indicator_show (indicator);
  
  g_hash_table_insert (indicator_context->indicators_by_name, g_strdup (name), indicator);
}

static void
handle_method_call (GDBusConnection *connection,
		    const gchar *sender,
		    const gchar *object_path,
		    const gchar *interface_name,
		    const gchar *method_name,
		    GVariant *parameters,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UnityWebappsIndicatorContext *context;
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  if (g_strcmp0 (method_name, "ShowIndicator") == 0)
    {
      const gchar *name;
      
      g_variant_get (parameters, "(s)", &name, NULL);
      
      real_show_indicator (context, name);

      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else if (g_strcmp0 (method_name, "ClearIndicator") == 0)
    {
      const gchar *name;
      
      g_variant_get (parameters, "(s)", &name, NULL);
      
      real_clear_indicator (context, name);
      
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else if (g_strcmp0 (method_name, "SetProperty") == 0)
    {
      const gchar *name;
      const gchar *property;
      const gchar *value;
      
      g_variant_get (parameters, "(sss)", &name, &property, &value, NULL);
      
      real_set_property (context, name, property, value);
      
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else
    {
      g_warning("Unexpected method in indicator context: %s", method_name);
      return;
    }
}
		  

static const GDBusInterfaceVTable interface_vtable =
  {
    handle_method_call,
    NULL,
    NULL
  };

UnityWebappsIndicatorContext *
unity_webapps_indicator_context_new (GDBusConnection *connection, const gchar *desktop_name)
{
  UnityWebappsIndicatorContext *indicator_context;
  GError *error;
  gchar *path, *type;
  
  error = NULL;
  
  if (indicator_introspection_data == NULL)
    {
      indicator_introspection_data = g_dbus_node_info_new_for_xml (_webapps_indicator, &error);

      if (error != NULL)
	{
	  g_error ("Can not load indicator introspection data: %s", error->message);
	  g_error_free (error);
	  
	  return NULL;
	}
      indicate_server_ref_default (); // HACK (from racarr) libindicate race condition bug.
    }
  
  indicator_context = g_malloc0 (sizeof (UnityWebappsIndicatorContext));
  
  indicator_context->connection = connection;
  
  indicator_context->object_registration 
    = g_dbus_connection_register_object (connection, 
					 UNITY_WEBAPPS_INDICATOR_PATH,
					 indicator_introspection_data->interfaces[0],
					 &interface_vtable, 
					 indicator_context,
					 NULL,
					 &error);
  
  if (error != NULL)
    {
      g_error ("Error registering indicator context object: %s", error->message);

      g_error_free (error);
      return NULL;
    }
  
  path = g_strdup_printf("%s/.data/applications/%s", g_get_home_dir(), desktop_name);

  indicator_context->indicate_server = g_object_new (INDICATE_TYPE_SERVER, "path", "/com/canonical/Unity/WebApps/Context/IndicateServer", NULL);
  
  //TODO: Fixme
  indicate_server_set_type(indicator_context->indicate_server, "message.im");
  indicate_server_set_desktop_file(indicator_context->indicate_server, path);
  
  indicate_server_show (indicator_context->indicate_server);
  
  indicator_context->indicators_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
  
  g_free (path);
			 
  
  return indicator_context;
}

void
unity_webapps_indicator_context_free (UnityWebappsIndicatorContext *indicator_context)
{
  g_hash_table_destroy (indicator_context->indicators_by_name);
  g_dbus_connection_unregister_object (indicator_context->connection, indicator_context->object_registration);

  g_object_unref (G_OBJECT (indicator_context->connection));
  g_object_unref (G_OBJECT (indicator_context->indicate_server));

  g_free (indicator_context);
}
