/*
 * unity-webapps-music_player-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include "unity-webapps-music-player-context.h"
#include "../gen-webapps-context-daemon.xml.h"
#include "../gen-webapps-service.xml.h"
#include "../gen-webapps-music-player.xml.h"

#include "../unity-webapps-dbus-defs.h"

#include "unity-webapps-utils.h"

static GDBusNodeInfo *music_player_introspection_data = NULL;

static void
real_set_track (UnityWebappsMusicPlayerContext *music_player_context,
		const gchar *artist,
		const gchar *album,
		const gchar *title,
		const gchar *art_location)
{
  UnityTrackMetadata  *metadata;
  GFile *artwork_file = NULL;
  
  metadata = unity_track_metadata_new ();
  unity_track_metadata_set_artist (metadata, artist);
  unity_track_metadata_set_album (metadata, album);
  unity_track_metadata_set_title (metadata, title);
  
  if (art_location != NULL)
    {
      artwork_file = g_file_new_for_path (art_location);
      unity_track_metadata_set_art_location (metadata, artwork_file);
    }
  
  unity_music_player_set_current_track (music_player_context->music_player, metadata);
  
  if (artwork_file != NULL)
    {
      g_object_unref (G_OBJECT (artwork_file));
    }
  g_object_unref (G_OBJECT (metadata));
}

typedef struct _unity_webapps_music_player_download_data {
  UnityWebappsMusicPlayerContext *context;
  gchar *artist;
  gchar *album;
  gchar *title;

} unity_webapps_music_player_download_data;

static void
_music_player_thumbnail_downloaded (gchar *tmp_name,
				    gpointer user_data)
{
  unity_webapps_music_player_download_data *ddata = (unity_webapps_music_player_download_data *)user_data;
  
  real_set_track (ddata->context, ddata->artist, ddata->album, ddata->title, tmp_name);
  
  g_free (ddata->artist);
  g_free (ddata->album);
  g_free (ddata->title);
  g_free (ddata);
  g_free (tmp_name);

}

static void
player_raise_callback (UnityMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerContext *music_player_context;
  GError *error;
  
  music_player_context = (UnityWebappsMusicPlayerContext *)user_data;
  
  error = NULL;
  g_dbus_connection_emit_signal (music_player_context->connection,
				 NULL,
				 UNITY_WEBAPPS_MUSIC_PLAYER_PATH,
				 UNITY_WEBAPPS_MUSIC_PLAYER_IFACE,
				 "Raise",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_warning ("Error emiting raise signal in music player context: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

static void
player_play_pause_callback (UnityMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerContext *music_player_context;
  GError *error;
  
  music_player_context = (UnityWebappsMusicPlayerContext *)user_data;
  
  error = NULL;
  g_dbus_connection_emit_signal (music_player_context->connection,
				 NULL,
				 UNITY_WEBAPPS_MUSIC_PLAYER_PATH,
				 UNITY_WEBAPPS_MUSIC_PLAYER_IFACE,
				 "PlayPause",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_warning ("Error emiting PlayPause signal in music player context: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

static void
player_previous_callback (UnityMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerContext *music_player_context;
  GError *error;
  
  music_player_context = (UnityWebappsMusicPlayerContext *)user_data;
  
  error = NULL;
  g_dbus_connection_emit_signal (music_player_context->connection,
				 NULL,
				 UNITY_WEBAPPS_MUSIC_PLAYER_PATH,
				 UNITY_WEBAPPS_MUSIC_PLAYER_IFACE,
				 "Previous",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_warning ("Error emiting previous signal in music player context: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

static void
player_next_callback (UnityMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerContext *music_player_context;
  GError *error;
  
  music_player_context = (UnityWebappsMusicPlayerContext *)user_data;
  
  error = NULL;
  g_dbus_connection_emit_signal (music_player_context->connection,
				 NULL,
				 UNITY_WEBAPPS_MUSIC_PLAYER_PATH,
				 UNITY_WEBAPPS_MUSIC_PLAYER_IFACE,
				 "Next",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_warning ("Error emiting next signal in music player context: %s", error->message);
      g_error_free (error);
      
      return;
    }
}


static void
handle_method_call (GDBusConnection *connection,
		    const gchar *sender,
		    const gchar *object_path,
		    const gchar *interface_name,
		    const gchar *method_name,
		    GVariant *parameters,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UnityWebappsMusicPlayerContext *context;
  
  context = (UnityWebappsMusicPlayerContext *)user_data;
  
  if (g_strcmp0 (method_name, "Init") == 0)
    {
      const gchar *title;
      
      g_variant_get (parameters, "(s)", &title, NULL);

      if (context->music_player == NULL)
	{
	  context->music_player = unity_music_player_new (context->desktop_name);
	  g_signal_connect (context->music_player, "raise", G_CALLBACK (player_raise_callback), context);
	  g_signal_connect (context->music_player, "play_pause", G_CALLBACK (player_play_pause_callback), context);
	  g_signal_connect (context->music_player, "next", G_CALLBACK (player_next_callback), context);
	  g_signal_connect (context->music_player, "previous", G_CALLBACK (player_previous_callback), context);
	}
      
      unity_music_player_set_title (context->music_player, title);
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else if (g_strcmp0 (method_name, "SetTrack") == 0)
    {
      const gchar *artist;
      const gchar *album;
      const gchar *title;
      const gchar *icon_url;
      
      g_variant_get (parameters, "(ssss)", &artist, &album, &title, &icon_url,  NULL);
      
      if (g_strcmp0(icon_url, "") == 0)
	{
	  real_set_track (context, artist, album, title, NULL);
	}
      else
	{
	  unity_webapps_music_player_download_data *ddata = (unity_webapps_music_player_download_data *) user_data;
	  
	  ddata = g_malloc (sizeof (unity_webapps_music_player_download_data));
	  
	  ddata->context = context;
	  ddata->artist = g_strdup (artist);
	  ddata->title = g_strdup (title);
	  ddata->album = g_strdup (album);
	  
	  unity_webapps_utils_tmp_file_from_url (icon_url, _music_player_thumbnail_downloaded, ddata);
	}
      
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else
    {
      g_warning("Unexpected method in music player context: %s", method_name);
      return;
    }
}

static GVariant *
handle_get_property (GDBusConnection *connection,
		     const gchar *sender,
		     const gchar *object,
		     const gchar *interface,
		     const gchar *property,
		     GError **error,
		     gpointer user_data)
{
  UnityWebappsMusicPlayerContext *context;
  
  context = (UnityWebappsMusicPlayerContext *)user_data;
  
  if (context->music_player == NULL)
    {
      return NULL;
    }
  
  if (g_strcmp0 (property, "CanGoPrevious") == 0)
    {
      return g_variant_new_boolean (unity_music_player_get_can_go_previous (context->music_player));
    }
  else if (g_strcmp0 (property, "CanGoNext") == 0)
    {
      return g_variant_new_boolean (unity_music_player_get_can_go_next (context->music_player));
    }
  else if (g_strcmp0 (property, "CanPlay") == 0)
    {
      return g_variant_new_boolean (unity_music_player_get_can_play (context->music_player));
    }
  else if (g_strcmp0 (property, "CanPause") == 0)
    {
      return g_variant_new_boolean (unity_music_player_get_can_pause (context->music_player));
    }
  else
    {
      g_warning("Invalid property get %s in music player context", property);
      *error = NULL;
      return NULL;
    }
}
static gboolean
handle_set_property (GDBusConnection *connection,
		     const gchar *sender,
		     const gchar *object,
		     const gchar *interface,
		     const gchar *property,
		     GVariant *value,
		     GError **error,
		     gpointer user_data)
{
  UnityWebappsMusicPlayerContext *context;
  
  context = (UnityWebappsMusicPlayerContext *)user_data;
  
  if (g_strcmp0 (property, "CanGoPrevious") == 0)
    {
      gboolean can_go_previous = g_variant_get_boolean (value);
      
      unity_music_player_set_can_go_previous (context->music_player, can_go_previous);
    }
  else if (g_strcmp0 (property, "CanGoNext") == 0)
    {
      gboolean can_go_next = g_variant_get_boolean (value);
      
      unity_music_player_set_can_go_next (context->music_player, can_go_next);
    }
  else if (g_strcmp0 (property, "CanPlay") == 0)
    {
      gboolean can_play_pause = g_variant_get_boolean (value);
      
      unity_music_player_set_can_play (context->music_player, can_play_pause);
    }
  else if (g_strcmp0 (property, "CanPause") == 0)
    {
      gboolean can_play_pause = g_variant_get_boolean (value);
      
      unity_music_player_set_can_pause (context->music_player, can_play_pause);
    }
  else
    {
      g_warning ("Invalid property set (%s) in music player context", property);
      *error = NULL;
      return FALSE;
    }
  
  return TRUE;
}
		  

static const GDBusInterfaceVTable interface_vtable =
  {
    handle_method_call,
    handle_get_property,
    handle_set_property
  };



UnityWebappsMusicPlayerContext *
unity_webapps_music_player_context_new (GDBusConnection *connection, const gchar *desktop_name)
{
  UnityWebappsMusicPlayerContext *music_player_context;
  GError *error;
  
  error = NULL;
  
  if (music_player_introspection_data == NULL)
    {
      music_player_introspection_data = g_dbus_node_info_new_for_xml (_webapps_music_player, &error);

      if (error != NULL)
	{
	  g_error ("Can not load music player introspection data: %s", error->message);
	  g_error_free (error);
	  
	  return NULL;
	}
    }
  
  music_player_context = g_malloc0 (sizeof (UnityWebappsMusicPlayerContext));
  
  music_player_context->connection = g_object_ref (connection);
  
  music_player_context->object_registration 
    = g_dbus_connection_register_object (connection, 
					 UNITY_WEBAPPS_MUSIC_PLAYER_PATH,
					 music_player_introspection_data->interfaces[0],
					 &interface_vtable, 
					 music_player_context,
					 NULL,
					 &error);
  
  if (error != NULL)
    {
      g_error ("Error registering music player context object: %s", error->message);

      g_error_free (error);
      return NULL;
    }
  
  music_player_context->desktop_name = g_strdup (desktop_name);
  
  
  return music_player_context;
}

void
unity_webapps_music_player_context_free (UnityWebappsMusicPlayerContext *music_player_context)
{
  g_dbus_connection_unregister_object (music_player_context->connection, music_player_context->object_registration);
  
  g_free (music_player_context->desktop_name);

  g_object_unref (G_OBJECT (music_player_context->music_player));
  g_object_unref (G_OBJECT (music_player_context->connection));

  g_free (music_player_context);
}

