/*
 * unity-webapps-notification-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <libnotify/notify.h>

#include "unity-webapps-notification-context.h"
#include "../gen-webapps-context-daemon.xml.h"
#include "../gen-webapps-service.xml.h"
#include "../gen-webapps-notification.xml.h"

#include "../unity-webapps-dbus-defs.h"

static GDBusNodeInfo *notification_introspection_data = NULL;

static void
real_show_notification (const gchar *summary,
			const gchar *body,
			const gchar *icon_url)
{
  NotifyNotification *notification;
  
  notification = notify_notification_new (summary, body, icon_url);
  notify_notification_show (notification, NULL);
  
  g_object_unref (G_OBJECT (notification));
}

/* Must free tmp_file */
static void
thumbnail_downloaded (gchar *tmp_file, gpointer user_data)
{
  gchar **arguments = (gchar **)user_data;
  
  real_show_notification (arguments[0], arguments[1], tmp_file);
  
  g_strfreev(arguments);
  g_free (tmp_file);
}

static void
show_notification (const gchar *summary,
		   const gchar *body,
		   const gchar *icon_url)
{
  if ((icon_url == NULL) || (g_strcmp0(icon_url, "") == 0))
    {
      real_show_notification (summary, body, NULL);
    }
  else
    {
      gchar **arguments;
      gboolean downloading_url;
      
      arguments = g_malloc(3*sizeof(gchar *));
      arguments[0] = g_strdup (summary);
      arguments[1] = g_strdup(body);
      arguments[2] = NULL;
      
      downloading_url = unity_webapps_utils_tmp_file_from_url (icon_url, thumbnail_downloaded, arguments);
      
      if (downloading_url == FALSE)
	{
	  g_strfreev (arguments);
	  real_show_notification (summary, body, NULL);
	}
      
    }
}

static void
handle_method_call (GDBusConnection *connection,
		    const gchar *sender,
		    const gchar *object_path,
		    const gchar *interface_name,
		    const gchar *method_name,
		    GVariant *parameters,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UnityWebappsNotificationContext *context;
  
  context = (UnityWebappsNotificationContext *)user_data;
  
  if (g_strcmp0 (method_name, "ShowNotification") == 0)
    {
      const gchar *summary;
      const gchar *body;
      const gchar *icon_url;
      
      g_variant_get (parameters, "(sss)", &summary, &body, &icon_url);
      
      show_notification (summary, body, icon_url);
       
      g_dbus_method_invocation_return_value (invocation, NULL);
    }
  else
    {
      g_warning("Unexpected method in notification context: %s", method_name);
      return;
    }
}
		  

static const GDBusInterfaceVTable interface_vtable =
  {
    handle_method_call,
    NULL,
    NULL
  };

UnityWebappsNotificationContext *
unity_webapps_notification_context_new (GDBusConnection *connection, const gchar *name)
{
  UnityWebappsNotificationContext *notification_context;
  GError *error;
  
  error = NULL;
  
  if (notification_introspection_data == NULL)
    {
      notification_introspection_data = g_dbus_node_info_new_for_xml (_webapps_notification, &error);

      if (error != NULL)
	{
	  g_error ("Can not load notification introspection data: %s", error->message);
	  g_error_free (error);
	  
	  return NULL;
	}
      
      notify_init (name);
    }
  
  notification_context = g_malloc0 (sizeof (UnityWebappsNotificationContext));
  
  notification_context->connection = connection;
  
  notification_context->object_registration 
    = g_dbus_connection_register_object (connection, 
					 UNITY_WEBAPPS_NOTIFICATION_PATH,
					 notification_introspection_data->interfaces[0],
					 &interface_vtable, 
					 notification_context,
					 NULL,
					 &error);
  
  if (error != NULL)
    {
      g_error ("Error registering notification context object: %s", error->message);

      g_error_free (error);
      return NULL;
    }
  
  return notification_context;
}

void
unity_webapps_notification_context_free (UnityWebappsNotificationContext *notification_context)
{
  g_dbus_connection_unregister_object (notification_context->connection, notification_context->object_registration);
  g_object_unref (G_OBJECT (notification_context->connection));

  g_free (notification_context);
}
