/*
 * unity-webapps-utils.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>
#include <stdlib.h>

#include <glib/gstdio.h>

#include "unity-webapps.h"

#define MAXIMUM_DOWNLOAD_SIZE 100*1024


typedef struct _unity_webapps_utils_download_data {
  UnityWebappsDownloadCallback callback;
  gpointer user_data;
  // TODO: Review the ownership of location?
  gchar *location;
} unity_webapps_utils_download_data;

void _download_streams_spliced (GObject *source_object,
		       GAsyncResult *res,
		       gpointer user_data)
{
  GOutputStream *ostream;
  gsize length;
  unity_webapps_utils_download_data *ddata;
  
  ddata = (unity_webapps_utils_download_data *)user_data;
  
  ostream = G_OUTPUT_STREAM (source_object);
  length = g_output_stream_splice_finish (ostream, res, NULL);
  
  if (ddata->callback)
    {
      ddata->callback(ddata->location, ddata->user_data);
    }
  
  g_free (ddata);
}

/*
 * Download a file to a URL respecting a maximum download size and providing an async callback. */
static gboolean
unity_webapps_utils_download_file_from_url (gchar *location, const gchar *url, UnityWebappsDownloadCallback callback, gpointer user_data)
{
  GFile *url_file, *location_file;
  GFileInfo *url_info;
  GFileInputStream *stream;
  GFileOutputStream *ostream;
  unity_webapps_utils_download_data *ddata;
  GError *error;
  
  url_file = g_file_new_for_uri (url);
  
  error = NULL;

  url_info = g_file_query_info (url_file, "standard::size",
				G_FILE_QUERY_INFO_NONE, NULL, &error);
  
  if (error != NULL)
    {
      g_warning ("Error querying file info for %s", url);
      g_object_unref (G_OBJECT (url_file));
      
      return FALSE;
    }

  if (g_file_info_get_size (url_info) > MAXIMUM_DOWNLOAD_SIZE) 
    {
      g_object_unref (G_OBJECT (url_info));
      g_object_unref (G_OBJECT (url_file));
      
      return FALSE;
    }
  
  stream = g_file_read (url_file, NULL, NULL);
  
  if (stream == NULL)
    {
      g_object_unref (G_OBJECT (url_file));
      return FALSE;
    }
  
  location_file = g_file_new_for_path (location);
  
  ostream = g_file_append_to (location_file, G_FILE_CREATE_REPLACE_DESTINATION, NULL, NULL);
  
  ddata = g_malloc (sizeof (unity_webapps_utils_download_data));
  ddata->location = location;
  ddata->callback = callback;
  ddata->user_data = user_data;
  
  g_output_stream_splice_async (G_OUTPUT_STREAM (ostream), G_INPUT_STREAM (stream), G_OUTPUT_STREAM_SPLICE_CLOSE_SOURCE | G_OUTPUT_STREAM_SPLICE_CLOSE_TARGET,
				0, NULL, _download_streams_spliced, ddata);
  
  g_object_unref (G_OBJECT (stream));
  g_object_unref (G_OBJECT (ostream));
  g_object_unref (G_OBJECT (location_file));
  g_object_unref (G_OBJECT (url_file));
  
  return TRUE;
}

static const gchar *
_icon_url_get_extension (const gchar *url) 
{
  const gchar *extension;
  
  extension = url + strlen (url);
  
  while (extension >= url)
    {
      if (*extension == '.')
	return extension;
      extension--;
    }

  return NULL;
}

/*
 * Asynchronously save an icon file for app_name at domain, given a URL.
 */
/*gchar *
unity_webapps_utils_icon_file_from_url (const gchar *domain, const gchar *app_name, const gchar *icon_url, UnityWebappsDownloadCallback callback, gpointer user_data)
{

  const gchar *extension;
  gchar *icon_filename, *icondir, *iconname;
  
  icondir = unity_webapps_dirs_get_icon_dir();
  
  extension = _icon_url_get_extension(icon_url);
  iconname = g_strdup_printf("%s-%s%s", domain, app_name, extension);
  
  icon_filename = g_build_filename(icondir, iconname, NULL);
  
  unity_webapps_utils_download_file_from_url (g_strdup (icon_filename), icon_url, callback, user_data);
   
  g_free (icondir);
  g_free (iconname);
  
  return icon_filename;
}
*/

/*
 * Download URL to a temporary file asynchronously.
 */
gboolean
unity_webapps_utils_tmp_file_from_url (const gchar *url, UnityWebappsDownloadCallback callback, gpointer user_data)
{
  gchar *tmp_name;
  
  tmp_name = tempnam(NULL, "uwatf");
  
  return unity_webapps_utils_download_file_from_url(tmp_name, url, callback, user_data);
  
}



