/*
 * unity-webapps-service.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-context.h"

UnityWebappsContext *
unity_webapps_context_new (UnityWebappsService *service, const gchar *name, const gchar *domain)
{
  GError *error = NULL;
  gchar *context_name;
  UnityWebappsContext *context;
  
  g_return_val_if_fail (service != NULL, NULL);
  
  unity_webapps_gen_service_call_get_context_sync (service->service_proxy, name, domain, &context_name, NULL /* Cancellable */, &error);
  
  if (error != NULL)
    {
      g_critical ("Error calling GetContext method when constructing new context proxy: %s", error->message);
      g_error_free (error);
      
      return NULL;
    }
  
  context = g_malloc0 (sizeof (UnityWebappsContext));
  
  context->context_proxy = 
    unity_webapps_gen_context_proxy_new_sync (service->session_bus,
					      G_DBUS_PROXY_FLAGS_NONE,
					      context_name,
					      UNITY_WEBAPPS_CONTEXT_PATH, 
					      NULL /* Cancellable */,
					      &error);
  
  if (error != NULL)
    {
      g_critical ("Error creating context proxy object for %s: %s", context_name, error->message);
      
      g_error_free (error);

      g_free (context_name);
      g_free (context);
      

      return NULL;
    }
  
  context->notification_context = unity_webapps_notification_context_new (service->session_bus, context_name);
  
  if (context->notification_context == NULL)
    {
      g_critical ("Failed to create notification context proxy");
      g_object_unref (G_OBJECT(context->context_proxy));
      g_free (context_name);
      g_free (context);
      
      return NULL;
    }
  
  context->indicator_context = unity_webapps_indicator_context_new (service->session_bus, context_name);
  
  if (context->indicator_context == NULL)
    {
      g_critical("Failed to create indicator context proxy");
      
      unity_webapps_notification_context_free (context->notification_context);
      g_object_unref (G_OBJECT (context->context_proxy));
      g_free (context_name);
      g_free (context);
      
      return NULL;	      
    }

  context->music_player_context = unity_webapps_music_player_context_new (service->session_bus, context_name);
  
  if (context->music_player_context == NULL)
    {
      g_critical("Failed to create music player context proxy");
      
      unity_webapps_notification_context_free (context->notification_context);
      unity_webapps_indicator_context_free (context->indicator_context);
      g_object_unref (G_OBJECT (context->context_proxy));
      g_free (context_name);
      g_free (context);
      
      return NULL;	      
    }
  
					      
  g_free (context_name);
  
  return context;
}

void 
unity_webapps_context_free (UnityWebappsContext *context)
{
  GError *error;
  
  error = NULL;
  
  unity_webapps_gen_context_call_lost_interest_sync (context->context_proxy,
						     NULL,
						     &error);
  
  if (error)
    {
      g_warning ("Error calling lost inteerest for context: %s",
		 error->message);
      
      g_error_free (error);
      
      error = NULL;
    }
  
  unity_webapps_notification_context_free (context->notification_context);
  unity_webapps_indicator_context_free (context->indicator_context);

  g_object_unref (G_OBJECT (context->context_proxy));
  
  g_free (context);
}

gchar *
unity_webapps_context_get_name (UnityWebappsContext *context)
{
  GError *error;
  gchar *name;
  
  error = NULL;
  
  unity_webapps_gen_context_call_get_name_sync (context->context_proxy,
						&name,
						NULL /* Cancellable */,
						&error);
  
  if (error != NULL)
    {
      g_critical ("Failure to invoke proxy method in unity_webapps_context_get_name: %s", error->message);
      g_error_free (error);
      
      return NULL;
    }
  
  return name;
}

gchar *
unity_webapps_context_get_domain (UnityWebappsContext *context)
{
  GError *error;
  gchar *domain;
  
  error = NULL;
  
  unity_webapps_gen_context_call_get_domain_sync (context->context_proxy,
						&domain,
						NULL /* Cancellable */,
						&error);
  
  if (error != NULL)
    {
      g_critical ("Failure to invoke proxy method in unity_webapps_context_get_domain: %s", error->message);
      g_error_free (error);
      
      return NULL;
    }
  
  return domain;
}

gchar *
unity_webapps_context_get_desktop_name (UnityWebappsContext *context)
{
  GError *error;
  gchar *desktop_name;
  
  error = NULL;
  
  unity_webapps_gen_context_call_get_desktop_name_sync (context->context_proxy,
						&desktop_name,
						NULL /* Cancellable */,
						&error);
  
  if (error != NULL)
    {
      g_critical ("Failure to invoke proxy method in unity_webapps_context_get_desktop_name: %s", error->message);
      g_error_free (error);
      
      return NULL;
    }
  
  return desktop_name;
}
