/*
 * unity-webapps-music-player-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-music-player-context.h"

UnityWebappsMusicPlayerContext *
unity_webapps_music_player_context_new (GDBusConnection *connection, const gchar *name)
{
  GError *error = NULL;
  UnityWebappsMusicPlayerContext *context;
  
  g_return_val_if_fail (connection != NULL, NULL);
  
  context = g_malloc0 (sizeof (UnityWebappsMusicPlayerContext));
  
  context->music_player_proxy = 
    unity_webapps_gen_music_player_proxy_new_sync (connection ,
						   G_DBUS_PROXY_FLAGS_NONE,
						   name,
						   UNITY_WEBAPPS_MUSIC_PLAYER_PATH, 
						   NULL /* Cancellable */,
						   &error);
  
  if (error != NULL)
    {
      g_critical ("Error creating music_player context proxy object for %s: %s", name, error->message);
      
      g_error_free (error);
      
      return NULL;
    }
  
  return context;
}

void 
unity_webapps_music_player_context_free (UnityWebappsMusicPlayerContext *context)
{
  g_object_unref (G_OBJECT (context->music_player_proxy));
  
  g_free (context);
}



void
unity_webapps_music_player_init (UnityWebappsContext *context,
				 const gchar *title)
{
  GError *error;
  
  error = NULL;
  
  unity_webapps_gen_music_player_call_init_sync (context->music_player_context->music_player_proxy,
						 title,
							      NULL /* Cancellable */,
							      &error);
  
  if (error)
    {
      g_critical ("Error invoking proxy method in unity_webapps_music_player_init: %s",
		  error->message);
      g_error_free (error);
      
      return;
    }
}

void 
unity_webapps_music_player_set_track (UnityWebappsContext *context,
				      const gchar *artist,
				      const gchar *album,
				      const gchar *title,
				      const gchar *icon_url)
{
  GError *error;
  
  error = NULL;
  
  unity_webapps_gen_music_player_call_set_track_sync (context->music_player_context->music_player_proxy,
						      artist,
						      album,
						      title,
						      icon_url ? icon_url : "",
							      NULL /* Cancellable */,
							      &error);
  
  if (error)
    {
      g_critical ("Error invoking proxy method in unity_webapps_music_player_set_track: %s",
		  error->message);
      g_error_free (error);
      
      return;
    }
}

static void
_do_callback (UnityWebappsGenMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerCallback callback = (UnityWebappsMusicPlayerCallback )user_data;
  
  if (callback)
    callback();
}

void 
unity_webapps_music_player_on_raise_callback (UnityWebappsContext *context, 
					      UnityWebappsMusicPlayerCallback callback)
{
  g_signal_connect (context->music_player_context->music_player_proxy, "raise", G_CALLBACK (_do_callback), callback);
}
void 
unity_webapps_music_player_on_play_pause_callback (UnityWebappsContext *context, 
						   UnityWebappsMusicPlayerCallback callback)
{
  g_signal_connect (context->music_player_context->music_player_proxy, "play_pause", G_CALLBACK (_do_callback), callback);
}
void 
unity_webapps_music_player_on_previous_callback (UnityWebappsContext *context, 
						 UnityWebappsMusicPlayerCallback callback)
{
  g_signal_connect (context->music_player_context->music_player_proxy, "previous", G_CALLBACK (_do_callback), callback);
}
void 
unity_webapps_music_player_on_next_callback (UnityWebappsContext *context, 
					     UnityWebappsMusicPlayerCallback callback)
{
  g_signal_connect (context->music_player_context->music_player_proxy, "next", G_CALLBACK (_do_callback), callback);
}

gboolean
unity_webapps_music_player_get_can_go_previous (UnityWebappsContext *context)
{
  return unity_webapps_gen_music_player_get_can_go_previous (context->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_go_next (UnityWebappsContext *context)
{
  return unity_webapps_gen_music_player_get_can_go_next (context->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_play (UnityWebappsContext *context)
{
  return unity_webapps_gen_music_player_get_can_play (context->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_pause (UnityWebappsContext *context)
{
  return unity_webapps_gen_music_player_get_can_pause (context->music_player_context->music_player_proxy);
}

void
unity_webapps_music_player_set_can_go_previous (UnityWebappsContext *context, gboolean can_go_previous)
{
  return unity_webapps_gen_music_player_set_can_go_previous (context->music_player_context->music_player_proxy, can_go_previous);
}

void
unity_webapps_music_player_set_can_go_next (UnityWebappsContext *context, gboolean can_go_next)
{
  return unity_webapps_gen_music_player_set_can_go_next (context->music_player_context->music_player_proxy, can_go_next);
}

void
unity_webapps_music_player_set_can_play (UnityWebappsContext *context, gboolean can_play)
{
  return unity_webapps_gen_music_player_set_can_play (context->music_player_context->music_player_proxy, can_play);
}

void
unity_webapps_music_player_set_can_pause (UnityWebappsContext *context, gboolean can_pause)
{
  return unity_webapps_gen_music_player_set_can_pause (context->music_player_context->music_player_proxy, can_pause);
}
