/*
 * unity-webapps-context-daemon.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-service.h"
#include "../gen-webapps-service.xml.h"
#include "../gen-webapps-context-daemon.xml.h"

#include "../unity-webapps-dbus-defs.h"

static GMainLoop *mainloop = NULL;

static GDBusNodeInfo *service_introspection_data = NULL;
static GDBusNodeInfo *context_introspection_data = NULL;

guint context_ready_signal_subscription;
guint context_lost_interest_signal_subscription;

UnityWebappsService *
unity_webapps_service_new ()
{
  UnityWebappsService *service;
  
  service = g_malloc0 (sizeof (UnityWebappsService));
  
  service->contexts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
  service->pending_ready = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
  service->lonely_daemons = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  
  return service;
}

void
unity_webapps_service_destroy (UnityWebappsService *service)
{
  g_hash_table_destroy (service->contexts);
  g_hash_table_destroy (service->pending_ready);
  g_hash_table_destroy (service->lonely_daemons);
  
  g_free (service);
}

static gboolean
spawn_context_daemon (const gchar *name,
		      const gchar *domain,
		      GError **error)
{
  gchar *commandline;
  gboolean ret;
  
  g_message ("Spawning context daemon %s %s", name, domain);
  
  
  commandline = g_strdup_printf (PACKAGE_LIBEXEC_DIR "/unity-webapps-context-daemon '%s' '%s'", name, domain);
  
  *error = NULL;
  ret = TRUE;
  
  g_spawn_command_line_async (commandline, error);
  
  if (*error != NULL)
    {
      g_critical("Error spawning context daemon: %s", (*error)->message);
      
      ret = FALSE;
      
    }
  
  g_free (commandline);
  
  return ret;
}

static void
add_interest (GDBusConnection *connection, 
	      UnityWebappsService *service, 
	      const gchar *context,
	      const gchar *sender)
{
  GDBusProxy *proxy;
  GError *error;
  GVariant *params;
  guint timeout_id;

  error = NULL;
  
  g_message ("Adding interest to existing daemon (%s) from %s", context, sender);

  // TODO: This could very well be async. Might be a race condition with killing the disinterested servers though...something to think about!
  proxy =
    g_dbus_proxy_new_sync (connection, G_DBUS_PROXY_FLAGS_NONE,
			   context_introspection_data->interfaces[0],
			   context,
			   UNITY_WEBAPPS_CONTEXT_PATH,
			   UNITY_WEBAPPS_CONTEXT_IFACE,
			   NULL /* Cancellable */,
			   &error);
  
  if (error != NULL)
    {
      g_critical ("Couldn't create proxy for interesting context (%s): %s", context, error->message);
      g_error_free (error);
      
      return;
    }
  
  params = g_variant_new ("(s)", sender);
  
  g_dbus_proxy_call_sync (proxy, "AddInterest",
			  params,
			  G_DBUS_CALL_FLAGS_NONE,
			  100,
			  NULL,
			  &error);

  if (error != NULL)
    {
      g_critical ("Failed to add interest to context (%s): %s", context, error->message);

      g_error_free (error);
      g_object_unref (G_OBJECT (proxy));

      return;
    }
  
  if ((timeout_id = GPOINTER_TO_UINT(g_hash_table_lookup(service->lonely_daemons, context))) != 0)
    {
      gboolean removed = FALSE;
      
      g_message ("Lonely daemon (%s) got interest from %s", context, sender);
      
      removed = g_source_remove (timeout_id);
      
      if (removed == FALSE)
	{
	  g_warning ("Couldn't remove kill timeout (%u) for %s", timeout_id, context);
	}
      
      g_hash_table_remove (service->lonely_daemons, context);
    }

  g_object_unref (G_OBJECT (proxy));
}

static void
handle_method_call (GDBusConnection *connection,
		    const gchar *sender,
		    const gchar *object_path,
		    const gchar *interface_name,
		    const gchar *method_name,
		    GVariant *parameters,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  
  UnityWebappsService *service = (UnityWebappsService *)user_data;
  
  if (g_strcmp0 (method_name, "HelloWorld") == 0)
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new("(s)", "Hello World"));
    }
  else if (g_strcmp0 (method_name, "GetContext") == 0)
    {
      GError *error;
      const gchar *domain;
      const gchar *name;
      const gchar *context;
      gchar *id;
      
      g_message ("Handling GetContext call for (%s,%s)", name, domain);

      
      g_variant_get(parameters,"(ss)", &name, &domain, NULL);
      
      id = g_strdup_printf("%s%s", name, domain);
      
      context = g_hash_table_lookup (service->contexts, id);
      
      if (context != NULL)
	{
	  g_message("Found existing context for (%s, %s): %s", name, domain, context);
	  add_interest (connection, service, context, sender);
	  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", context));
	}
      else
	{
	  g_message("No existing context for (%s, %s)", name, domain);
	  //. TODO: Make sure there isn't a race condition here
	  error = NULL;
	  spawn_context_daemon (name, domain, &error);

	  if (error != NULL)
	    {
	      g_warning("Couldn't spawn context daemon for %s (%s)", name, domain);
	      g_dbus_method_invocation_return_gerror (invocation, error);
	      g_error_free (error);
	    }
	  else
	    {
	      g_message ("Waiting for context (%s, %s) to send ContextReady", name, domain);
	      g_hash_table_insert (service->pending_ready, g_strdup(id), g_object_ref (invocation));
	    }
	}


      g_free (id);
    }
  else
    {
      g_warning ("Unidentified method call to interface %s, %s", interface_name, method_name);
    }
}

static const GDBusInterfaceVTable interface_vtable =
  {
    handle_method_call,
    NULL,
    NULL
  };

typedef struct _context_vanished_user_data {
  UnityWebappsService *service;
  guint watcher_id;
} context_vanished_user_data;

static void
context_name_vanished (GDBusConnection *connection,
		       const gchar *name,
		       gpointer user_data)
{
  context_vanished_user_data *data= (context_vanished_user_data *)user_data;
  GHashTableIter iter;
  gpointer key, value;
  guint timeout_id;
  
  g_message ("Watched context (%s) vanished", name);
  
  if ((timeout_id = GPOINTER_TO_UINT (g_hash_table_lookup (data->service->lonely_daemons, name))) != 0)
    {
      gboolean removed = FALSE;
      
      g_message("Lost context (%s) was slated to be killed, removing it's timeout", name);

      removed = g_source_remove (timeout_id);
      
      if (removed == FALSE)
	{
	  g_warning ("Failed to remove timeout_id (%u) when context vanished (%s)", timeout_id, name);
	}
      
      g_hash_table_remove (data->service->lonely_daemons, name);
    }
  
  g_hash_table_iter_init (&iter, data->service->contexts);
  
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      if (g_strcmp0(value, name) == 0)
	{
	  g_message("Unregistering context: %s", (gchar *)key);
	  g_hash_table_remove(data->service->contexts, key);
	  
	  g_bus_unwatch_name (data->watcher_id);
	  
	  return;
	}
    }
}

static void
clear_queue (GHashTable *pending_ready, GError *error)
{
  GHashTableIter iter;
  gpointer key, value;
  
  g_hash_table_iter_init (&iter, pending_ready);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      GDBusMethodInvocation *invocation = (GDBusMethodInvocation *)value;
      
      g_assert (invocation != NULL);
      
      g_dbus_method_invocation_return_gerror (invocation, error);
    }
  g_hash_table_remove_all (pending_ready);
}

static void
register_proxy (UnityWebappsService *service,
		GDBusConnection *connection,
		const gchar *sender,
		const gchar *object)
{
 GDBusMethodInvocation *invocation;
 GDBusProxy *proxy;
 GVariant *domainv, *namev;
 const gchar *domain;
 const gchar *name;
 gchar *id;
 GError *error;
 
 error = NULL;

 proxy = 
   g_dbus_proxy_new_sync (connection,
			  G_DBUS_PROXY_FLAGS_NONE,
			  context_introspection_data->interfaces[0],
			  sender,
			  object,
			  UNITY_WEBAPPS_CONTEXT_IFACE, 
			  NULL, /* Cancellable */
			  &error/* Error */);
 
 if (error != NULL)
   {
     g_message("Error creating context proxy (%s %s): %s", sender, object, error->message);
     
     clear_queue(service->pending_ready, error);
     
     g_error_free (error);
     
     return;
   }
 
 domainv = g_dbus_proxy_call_sync (proxy, "GetDomain",
				   NULL,
				   G_DBUS_CALL_FLAGS_NONE,
				   100,
				   NULL, /*Cancellable*/
				   &error /* Error */);
 if (error != NULL)
   {
     printf("Error calling GetDomain method of proxy: %s", error->message);
     
     clear_queue (service->pending_ready, error);

     g_error_free (error);
     
     return;
   }
 
 namev = g_dbus_proxy_call_sync (proxy, "GetName",
				 NULL,
				 G_DBUS_CALL_FLAGS_NONE,
				 100,
				 NULL /* Cancellable */,
				 &error /* Error */);
 
 if (error != NULL)
   {
     printf("Error calling GetName method of proxy: %s", error->message);
     
     clear_queue (service->pending_ready, error);

     g_error_free (error);
     g_variant_unref (domainv);

     return;
   }
 
 domain = g_variant_get_string (g_variant_get_child_value (domainv, 0), NULL);
 name = g_variant_get_string (g_variant_get_child_value (namev, 0), NULL);
 
 id = g_strdup_printf("%s%s", name, domain);
 
 
 invocation = (GDBusMethodInvocation *)g_hash_table_lookup(service->pending_ready, id);
 
 if (invocation != NULL)
   {

     GVariant *param = g_variant_new ("(s)",g_dbus_method_invocation_get_sender (invocation));

     g_message ("Context daemon (%s %s) is ready for %s at %s", name, domain, g_dbus_method_invocation_get_sender (invocation), sender);
     g_message("Registering interest in new context daemon (%s, %s)", name, domain);

     g_dbus_proxy_call_sync (proxy, "AddInterest",
			     param,
			     G_DBUS_CALL_FLAGS_NONE,
			     100,
			     NULL,
			     &error);
     
     if (error != NULL)
       {
	 g_critical ("Error calling AddInterest (%s): %s", g_dbus_method_invocation_get_sender (invocation), error->message);
	 g_error_free (error);
       }
     g_dbus_method_invocation_return_value (invocation, g_variant_new("(s)", sender));
     g_hash_table_remove (service->pending_ready, id);
   }
 
 g_hash_table_insert (service->contexts, id, g_strdup (sender));
 
 g_object_unref (G_OBJECT (proxy));

 g_variant_unref (domainv);
 g_variant_unref (namev);
			   
}

static void
bus_filter_context_ready (GDBusConnection *connection,
			  const gchar *sender,
			  const gchar *object,
			  const gchar *interface,
			  const gchar *signal,
			  GVariant *params,
			  gpointer user_data)
{
  UnityWebappsService *service = (UnityWebappsService *)user_data;
  context_vanished_user_data *data;
  guint watcher_id;
 

  if (g_strcmp0(signal, "ContextReady") != 0) 
    {
      g_warning ("Unexpected signal for webapps service: %s", signal);
      return;
    }
  
  data = g_malloc0 (sizeof (context_vanished_user_data));
  data->service = service;

  watcher_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
		    sender,
		    G_BUS_NAME_WATCHER_FLAGS_NONE,
		    NULL,
		    context_name_vanished,
		    data /* User data*/,
		    g_free /* User data free*/);
  
  data->watcher_id = watcher_id;
  
  register_proxy (service, connection, sender, object);
  
}

typedef struct _context_lost_interest_user_data {
  GDBusConnection *connection;
  gchar *sender;
  UnityWebappsService *service;
} context_lost_interest_user_data;

static void 
context_lost_interest_user_data_free (gpointer p)
{
  context_lost_interest_user_data *data = (context_lost_interest_user_data *)p;
  
  // Sender is managed by the hash table.
  g_object_unref(data->connection);
  
  g_free(data);
}

static gboolean
kill_lonely_daemon (gpointer user_data)
{
  context_lost_interest_user_data *data;
  GDBusProxy *proxy;
  GError *error;
  
  data = (context_lost_interest_user_data *)user_data;
  
  error = NULL;

  proxy = 
    g_dbus_proxy_new_sync (data->connection,
			   G_DBUS_PROXY_FLAGS_NONE,
			   context_introspection_data->interfaces[0],
			   data->sender,
			   UNITY_WEBAPPS_CONTEXT_PATH,
			   UNITY_WEBAPPS_CONTEXT_IFACE, 
			   NULL, /* Cancellable */
			   &error/* Error */);
  
  if (error != NULL)
    {
      g_warning ("Couldn't create proxy for lonely daemon (%s), treating it as dead: %s", data->sender, error->message);

      g_error_free (error);
      g_hash_table_remove(data->service->lonely_daemons, data->sender);
      
      return FALSE;
    }
  
  g_dbus_proxy_call_sync (proxy, "Shutdown",
			  NULL,
			  G_DBUS_CALL_FLAGS_NONE,
			  100,
			  NULL,
			  &error);
  
  if (error != NULL)
    {
      g_warning ("Failed to call shutdown on lonely daemon (%s), treating it as dead: %s", data->sender, error->message);

      g_error_free (error);
      g_object_unref (G_OBJECT (proxy));
      g_hash_table_remove(data->service->lonely_daemons, data->sender);
      
      return FALSE;
    }

  g_object_unref (G_OBJECT (proxy));
  g_hash_table_remove(data->service->lonely_daemons, data->sender);

  return FALSE;
}

static void
bus_filter_context_lost_interest (GDBusConnection *connection,
				  const gchar *sender,
				  const gchar *object,
				  const gchar *interface,
				  const gchar *signal,
				  GVariant *params,
				  gpointer user_data)
{
  UnityWebappsService *service = (UnityWebappsService *)user_data;
  context_lost_interest_user_data *data;
  guint timeout_id;
  
  g_message("Server (%s) has no interest", sender);
  
  data = g_malloc0 (sizeof (context_lost_interest_user_data));

  data->sender = g_strdup (sender);
  data->connection = g_object_ref(connection);
  data->service = service;
  
  timeout_id = g_timeout_add_seconds_full (G_PRIORITY_LOW, 60, kill_lonely_daemon, data, context_lost_interest_user_data_free);
  
  g_hash_table_insert (service->lonely_daemons, data->sender, GUINT_TO_POINTER(timeout_id));

  
}


static void
on_bus_acquired (GDBusConnection *connection,
		 const gchar *name,
		 gpointer user_data)
{
  UnityWebappsService *service;
  GError *error;
  guint reg_id;
  
  service = (UnityWebappsService *)user_data;
  
  error = NULL;

  reg_id = g_dbus_connection_register_object (connection, 
					      UNITY_WEBAPPS_SERVICE_PATH,
					      service_introspection_data->interfaces[0],
					      &interface_vtable,
					      service /* user data */,
					      NULL /* user data free */,
					      &error /* Error */);

  if (error != NULL)
    {
      g_error ("Can not register object %s: %s", UNITY_WEBAPPS_SERVICE_PATH, error->message);
      g_error_free (error);

      return;
    }
  
  context_ready_signal_subscription = 
    g_dbus_connection_signal_subscribe (connection,
					NULL,
					UNITY_WEBAPPS_CONTEXT_IFACE,
					"ContextReady",
					NULL,
					NULL,
					G_DBUS_SIGNAL_FLAGS_NONE,
					bus_filter_context_ready,
					service /* user data*/,
					NULL /* destroy notify */);

  context_lost_interest_signal_subscription = 
    g_dbus_connection_signal_subscribe (connection,
					NULL,
					UNITY_WEBAPPS_CONTEXT_IFACE,
					"NoInterest",
					NULL,
					NULL,
					G_DBUS_SIGNAL_FLAGS_NONE,
					bus_filter_context_lost_interest,
					service /* user data*/,
					NULL /* destroy notify */);
  
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_SERVICE_PATH,
				 UNITY_WEBAPPS_SERVICE_IFACE,
				 "ServiceReady",
				 NULL /* Params*/,
				 &error /* Error */);
  
  if (error != NULL)
    {
      g_error ("Can not emit signal ServiceReady on com.canonical.Unity.Webapps.Service: %s",
	       error->message);
      g_error_free (error);
      
      return;
    }
}

static void
on_name_acquired (GDBusConnection *connection,
		  const gchar *name,
		  gpointer user_data)
{
}

static void
on_name_lost (GDBusConnection *connection,
	      const gchar *name,
	      gpointer user_data)
{
  g_main_loop_quit (mainloop);
}

gint 
main (gint argc, gchar **argv)
{
  UnityWebappsService *service;
  GError *error;
  guint owner_id;

  g_type_init();
  
  service = unity_webapps_service_new ();
  
  error = NULL;
  
  service_introspection_data = g_dbus_node_info_new_for_xml (_webapps_service, &error);
  
  if (error != NULL)
    {
      g_error ("Error parsing service introspection data: %s", error->message);
      g_error_free (error);

      return 1;
    }
  g_assert (service_introspection_data != NULL);
  
  context_introspection_data = g_dbus_node_info_new_for_xml (_webapps_context_daemon, &error);
  if (error != NULL)
    {
      g_error ("Error parsing service introspection data: %s", error->message);
      g_error_free (error);

      return 1;
    }
  g_assert (context_introspection_data != NULL);
  
  owner_id = g_bus_own_name (G_BUS_TYPE_SESSION, UNITY_WEBAPPS_SERVICE_NAME,
			     G_BUS_NAME_OWNER_FLAGS_NONE,
			     on_bus_acquired,
			     on_name_acquired,
			     on_name_lost,
			     service /* userdata */,
			     NULL /* userdata free*/);
			     
  
  mainloop = g_main_loop_new(NULL, FALSE);
  g_main_loop_run(mainloop);
  
  g_bus_unown_name(owner_id);

  g_dbus_node_info_unref (service_introspection_data);
  g_dbus_node_info_unref (context_introspection_data);
  
  unity_webapps_service_destroy (service);

  return 0;
}
